/*
	FlashLoad.c
	
	This file builds a little genesis segment which is used
	to load flash images.  The cool thing is that flash segments
	now include 
	
	1.  Table of contents
	2.  File blocks
	
	The table of contents is a structure containing checksums for each of
	the following data blocks.  This will allow us to download it through
	the Psyq with better reliability.
	
	With this information, we'll be able to repeat blockreads in case
	they come down wrong (which they often do!)
	
	Large portions of this program were lifted from Shannon's
	original flashing tool.

	Andy Stadler
	Copyright (C) 1995	Captapult Entertainment, Inc.
	All Rights Reserved
*/

#include "FlashCard.h"
#include "asm.h"


//	Memory Map

//	0xFFF000			- program runs here
//	0xFFE000			- 4k read buffer
//  0xFFD000			- small buffer for the file header


#define kFlashBase		0x200000
#define	kHeaderBuf		0xFFD000
#define kReadBuf		0xFFE000

//	a few globals, up at the very top

#define	gBadReads		(-2)
#define kStackBase		(-16)




//	Register Map

//	A3	= Base Addr of Flash
//	A5	= File Header

//	D4	= error counter (used?)
//	D5	= file handle
//	D6	= current block #


//	INTEL flash commands

enum	{
	kResetArray		= 0xFF,
	kGetID1			= 0x90,		// cycle 2 is a read
	kReadSR1		= 0x70,		// cycle 2 is a read
	kClearSR		= 0x50,
	kErase1			= 0x20,
	kErase2			= 0xD0,
	kEraseSusp1		= 0xB0,
	kEraseSusp2		= 0xD0,
	kByteWrite1		= 0x40		// cycle 2 is write data
};

//	INTEL status bits

enum	{
	kSystemReady	= 0x80,
	kEraseSuspended	= 0x40,
	kEraseError		= 0x20,
	kWriteError		= 0x10,
	kPowerError		= 0x08,
	
	kEraseErrors	= kEraseError + kWriteError + kPowerError,
	kWriteErrors	= kWriteError + kPowerError
};


#define	kFileName		"ROMIMAGE.FLA"


#define	POLHOST		dc.w 0xA000
#define HALT		dc.w 0xA003


void	EraseAllBlocks( void );
void	EraseBlock( long address );
void	ReadHeader( void );
void	SetupPsyQ( void );
void	ClearSR( void );
void	WaitForReady( void );
void	CheckEraseError( void );
UShort	ccitt_updcrc_sega(UShort icrc, void* icp, unsigned short icnt);
void	ReadOneBlock( void );
void	FlashOneBlock( void );
void	PrintMessage( char* msg );
void	PrintShort( short num );
void	PrintCR( void );



void flash ( void );
void flash ( void )
{
	asm	{
		move.l	#kStackBase,sp
		jsr		SetupPsyQ
		
			// setup FRED to view the flash
			
		move.b	#4,0x3bfe03
		
			// ok, stop before starting
			
		dc.w	0xA005				// install vectors
		dc.w	0xA003				// halt
		
			//	clear all globals
			
		clr.w	gBadReads
		
			//	say "hello"
		
		pea		@WelcomeMessage
		jsr		PrintMessage
		addq.l	#4,sp
		
			//	setup registers, begin erase cycles
			
		clr.l	d5					// clear file pointer
		move.l	#kFlashBase,A3
		
		jsr		EraseAllBlocks
		
		pea		@ErasedMessage
		jsr		PrintMessage
		addq.l	#4,sp
		
		move.w	#0xFF,(a3)			// reset flash to read array mode
		
			// open source file
			
		move.w	#3,d4				// init error counter
		
		dc.w	0xA201				// init file services
@ReOpen	
		lea		@FileName,a0		// A0 = filename string
		clr.w	d1					// D1 = 0 = readonly access
		dc.w	0xA203				// OPEN()
		beq.s	@OpenOK
		
		dbra	d4,@ReOpen			// try again a few times?
		bra		@BailOut
@OpenOK
		move.w	d1,d5				// save the file handle
		
		jsr		ReadHeader			// read the header
		
			// now, start reading & writing....
			
		move.l	#kFlashBase,A3		// trashed by OPEN
		clr.w	d6					// start from block zero

@MainReadLoop		
		jsr		ReadOneBlock
		jsr		FlashOneBlock
		
		addq.w	#1,d6
		cmp.w	FileHeader.numBlocks(a5),d6
		bne.s	@MainReadLoop
		
		
			// All done!  print out the stats of the flash session
			
		pea		@Flashed1Message
		jsr		PrintMessage
		addq.l	#4,sp
		pea		@Flashed2Message
		jsr		PrintMessage
		addq.l	#4,sp
		move.w	gBadReads,-(sp)
		jsr		PrintShort
		addq.l	#2,sp
		jsr		PrintCR
		pea		@Flashed3Message
		jsr		PrintMessage
		addq.l	#4,sp
		move.w	FileHeader.numBlocks(a5),-(sp)
		jsr		PrintShort
		addq.l	#2,sp
		jsr		PrintCR

@BailOut
		move.w	#0xFF,(a3)			// reset flash mode
		
		move.w	d5,d1				// close the file
		beq.s	@NoClose			//  (unless it was never opened)
		dc.w	0xA204				// CLOSE()
@NoClose

		dc.w	0xA003				// done!
		


@FileName
		dc.b	kFileName,0
@WelcomeMessage
		dc.b	"\r\r---FLASH PROGRAMMING---\rBeginning Erase Cycle\r",0
@ErasedMessage
		dc.b	"Erase was successful\r",0

@Flashed1Message
		dc.b	"FLASH COMPLETE\r",0
@Flashed2Message
		dc.b	"Bad reads:    ",0
@Flashed3Message
		dc.b	"Total blocks: ",0
		
	}
	
}



void SetupPsyQ( void )
{
	asm	{
		move.b	0x3f204f,d0
		cmp.b	#0x32,d0
		bne.s	@try202
		move.l	#0x3f15CC,0x200028			// v2.05
		bra.s	@Exit

@try202	// it wasn't a 2.05, so try 2.02
		move.b	0x3f1ec1,d0
		cmp.b	#0x32,d0
		bne.s	@do204
		move.l	#0x3f1476,0x200028			// v2.02
		bra.s	@Exit

@do204	// wasnt a 2.02 either so must be a 2.04						
		move.l	#0x3f1456,0x200028			// v2.04
@Exit
	}
}		// rts



//-----------------------------------------------------------------------
//	FLASH CHIP SUBROUTINES
//-----------------------------------------------------------------------


	// Note the funny block sizes.
	//	A 512k flash chip is made from the following blocks:
	//		16, 8, 8, 96, 128, 128, 128
	
void EraseAllBlocks( void )
{
	asm	{
		pea		0x200000
		bsr		EraseBlock
		addq.w	#4,sp

		pea		0x204000
		bsr		EraseBlock
		addq.w	#4,sp

		pea		0x206000
		bsr		EraseBlock
		addq.w	#4,sp

		pea		0x208000
		bsr		EraseBlock
		addq.w	#4,sp

		pea		0x220000
		bsr		EraseBlock
		addq.w	#4,sp

		pea		0x240000
		bsr		EraseBlock
		addq.w	#4,sp

		pea		0x260000
		bsr		EraseBlock
		addq.w	#4,sp
	}
}	// rts


void EraseBlock( long address )
{
	asm	{
		jsr		ClearSR
		move.l	address,a0
		
		move.w	#kErase1,(a0)		// erase cmd 1
		move.w	#kErase2,(a0)		// erase cmd 2
		
		jsr		WaitForReady
		jsr		CheckEraseError
		
		move.w	#kResetArray,(a0)
	}
}	// rts



void ClearSR( void )
{
	asm	{
		move.w	#kClearSR,(a3)
	}
}	// rts


void WaitForReady( void )
{
	asm	{
@Loop
		move.w	#kReadSR1,(a3)		// request SR
		move.w	(a3),d0				// read it
		move.b	d0,d0				// sample bit 7
		bpl.s	@Loop				// 1 = ready, 0 = busy
	}
}	// rts


void CheckEraseError( void )
{
	asm	{
		move.w	#kReadSR1,(a3)		// request SR
		move.w	(a3),d0				// read it
		and.w	#kEraseErrors,d0	// any erase error bits set?
		beq.s	@ExitOK
		
		dc.w	0xA003		// HALT - erase error!
@ExitOK
	}
}	// rts




	// flash block # D6

void	FlashOneBlock( void )
{
	ClearSR();
	
	asm	{
		move.w	d6,d0
		mulu.w	#kBlockSize,d0
		move.l	d0,a0
		add.l	a3,a0			// a0 = dest addr
		move.l	#kReadBuf,a1	// a1 = src addr
		
		move.w	#kBlockSize,d1	// d1 = counter
		
@ProgramLoop

		jsr		WaitForReady
		
		move.w	#kByteWrite1,(a0)	// setup for write
		move.w	(a1)+,(a0)+			// write the data
	
		subq.w	#2,d1				// done?
		bne.s	@ProgramLoop
		
			// final checks, did the programming take?
				
		
		move.w	#kReadSR1,(a3)
		move.w	(a3),d0
		and.w	#kWriteErrors,d0
		beq.s	@CheckWrittenData
		
		pea		@BadFlashBlock
		jsr		PrintMessage
		addq.l	#4,sp
		
		dc.w	0xA003		// STOP - can't flash a block
		
			// flash seems to have completed - checksum it!
		
@CheckWrittenData
		
		jsr		WaitForReady
		move.w	#0xFF,(a3)			// reset flash mode

		move.w	d6,d0
		mulu.w	#kBlockSize,d0
		move.l	d0,a0
		add.l	a3,a0			// a0 = dest addr

		move.w	#kBlockSize,-(sp)
		pea		(a0)
		move.w	#0xFFFF,-(sp)
		jsr		ccitt_updcrc_sega
		addq.l	#2+4+2,sp

			// look up the checksum in the table
			
		move.w	d6,d1
		add.w	d1,d1					// table of words
		add.w	FileHeader.skipToChecks(a5),d1
		cmp.w	(a5,d1.w),d0
		beq.s	@GoodBlockWritten
		
		pea		@BadFlashCheck
		jsr		PrintMessage
		addq.l	#4,sp
		
		dc.w	0xA003				// HALT - flash data doesn't look right
		
		
@BadFlashBlock
		dc.b	"Error in flash SR\r",0
@BadFlashCheck
		dc.b	"Error in flash checksum\r",0
		

@GoodBlockWritten
	}
}	// rts




//-----------------------------------------------------------------------
//	FILE READ SUBROUTINES
//-----------------------------------------------------------------------


void ReadHeader( void )
{
	asm	{
	
		move.l	#kHeaderBuf,a5			// setup A5 to the special header
	
			// First step:  Ensure we can read the file header
		
@ReadLoop
		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!
		
		move.w	d5,d1					// file handle
		clr.l	d2						// position = 0
		clr.w	d3						// seek mode = 0 (absolute)
		dc.w	0xA207					// SEEK()
		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!

		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!

		move.w	d5,d1					// file handle
		move.w	#sizeof(FileHeader),d2	// byte count
		move.l	a5,a0					// data buffer
		dc.w	0xA205					// READ()
		
		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!

			//	Q:  Did we get a good header?
			
		tst.w	d0							// did READ() report an error?
		bne.s	@BadHeader
		cmp.w	#sizeof(FileHeader),d2		// did we read enough bytes?
		bne.s	@BadHeader
		cmp.w	#1,FileHeader.version(a5)	// version = 1?
		bne.s	@BadHeader
		cmp.w	#kHeaderSig,FileHeader.signature(a5)
		
		move.w	FileHeader.headerCksum(a5),-(sp)
		clr.w	FileHeader.headerCksum(a5)
		
		move.w	#sizeof(FileHeader),-(sp)
		pea		(a5)
		move.w	#0xFFFF,-(sp)
		jsr		ccitt_updcrc_sega
		addq.l	#2+4+2,sp
		
		cmp.w	(sp)+,d0				// header checksum ok?
		beq.s	@GoodHeader
		
@BadHeader
		dbra	d4,@ReadLoop			// try again a few times?

	
			// Second step:  Ensure we can read the checksum table			
		
@GoodHeader		
@ReadChecksLoop	
		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!
		
		move.w	d5,d1					// file handle
		clr.l	d2						// position = after header
		add.w	FileHeader.skipToChecks(a5),d2
		clr.w	d3						// seek mode = 0 (absolute)
		dc.w	0xA207					// SEEK()

		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!
		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!

		move.w	d5,d1					// file handle
		move.w	FileHeader.numBlocks(a5),d2	// byte count = 2*numBlocks
		add.w	d2,d2
		move.l	a5,a0					// dest = after header
		add.w	FileHeader.skipToChecks(a5),a0
		dc.w	0xA205					// READ()
		
		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!

			//	Q:  Did we get good checksums?
			
		move.w	FileHeader.numBlocks(a5),d0	// byte count = 2*numBlocks
		add.w	d0,d0
		move.w	d0,-(sp)
		move.l	a5,a0						// dest = after header
		add.w	FileHeader.skipToChecks(a5),a0
		move.l	a0,-(sp)
		move.w	#0xFFFF,-(sp)
		jsr		ccitt_updcrc_sega
		addq.l	#2+4+2,sp
		
		cmp.w	FileHeader.checksCksum(a5),d0
		beq.s	@GoodChecks
		
		dbra	d4,@ReadChecksLoop			// try again a few times?
		
@GoodChecks
	}
}	// rts


	// read the block # from D6, and repeat until we get it right
	
	//	NOTE:  There is some serious flakiness in reading 4k blocks,
	//	so I'm going to do a single seek and 4 1k reads.

void	ReadOneBlock( void )
{
	asm	{
@ReadBlockLoop

			// seek to the start of the 4k chunk

		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!
		
		move.w	d5,d1					// file handle
		move.w	d6,d2					// position = blocks + numBlocks*size
		mulu.w	#kBlockSize,d2			// now we have a long
		add.w	FileHeader.skipToData(a5),d2
		clr.w	d3						// seek mode = 0 (absolute)
		dc.w	0xA207					// SEEK()

		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!

			// perform four sequential 1k reads to get the data

		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!

		move.w	d5,d1					// file handle
		move.w	#kBlockSize/4,d2			// byte count = 1/4 block
		move.l	#kReadBuf+0x0000,a0			// dest = read buffer
		dc.w	0xA205					// READ()
		
		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!

		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!

		move.w	d5,d1					// file handle
		move.w	#kBlockSize/4,d2			// byte count = 1/4 block
		move.l	#kReadBuf+0x0400,a0			// dest = read buffer
		dc.w	0xA205					// READ()
		
		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!

		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!

		move.w	d5,d1					// file handle
		move.w	#kBlockSize/4,d2			// byte count = 1/4 block
		move.l	#kReadBuf+0x0800,a0			// dest = read buffer
		dc.w	0xA205					// READ()
		
		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!

		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!

		move.w	d5,d1					// file handle
		move.w	#kBlockSize/4,d2			// byte count = 1/4 block
		move.l	#kReadBuf+0x0C00,a0			// dest = read buffer
		dc.w	0xA205					// READ()
		
		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!


			// compute the new block's checksum
			
		move.w	#kBlockSize,-(sp)
		pea		kReadBuf
		move.w	#0xFFFF,-(sp)
		jsr		ccitt_updcrc_sega
		addq.l	#2+4+2,sp

			// look up the checksum in the table
			
		move.w	d6,d1
		add.w	d1,d1					// table of words
		add.w	FileHeader.skipToChecks(a5),d1
		cmp.w	(a5,d1.w),d0
		beq.s	@GoodBlockRead
		
		pea		@BadBlockMessage
		jsr		PrintMessage
		addq.l	#4,sp

		addq.w	#1,gBadReads
		
		bra		@ReadBlockLoop

@BadBlockMessage
		dc.b	"Bad block was read\r",0
		
@GoodBlockRead
	}
}	// rts


	// use "stdio" on the debugger
	
void PrintMessage( char* msg )
{
char*	scan = msg;
short	length;

	while (*scan++) ;
	length = ( scan - msg ) - 1;	// don't pass the zero
	
	asm	{
		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!		
		POLHOST
		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!

		movem.l	d3-d7/a1-a5,-(sp)		// in case PsyQ blasts them!

		move.w	#-1,d1				// -1 is "message window 1"
		move.w	length,d2
		move.l	msg,a0
		dc.w	0xA206
		
		POLHOST

		movem.l	(sp)+,d3-d7/a1-a5		// in case PsyQ blasted them!
	}
}

void	PrintShort( short num )
{
char	string[6];

	string[0] = ((num >> 24) & 0xF) + '0';
	if (string[0] > '9')
		string[0] += ('A' - ('9' + 1));
	string[1] = ((num >> 16) & 0xF) + '0';
	if (string[1] > '9')
		string[1] += ('A' - ('9' + 1));
	string[2] = ((num >>  8) & 0xF) + '0';
	if (string[2] > '9')
		string[2] += ('A' - ('9' + 1));
	string[3] = ((num >>  0) & 0xF) + '0';
	if (string[3] > '9')
		string[3] += ('A' - ('9' + 1));
	string[4] = 0;
	
	PrintMessage(string);
}

void	PrintCR( void )
{
	asm	{
		pea		@CRMessage
		jsr		PrintMessage
		addq.l	#4,sp
		rts
	
@CRMessage
		dc.b	'\r',0
	}
}





	// cribbed the good ol' CRC routine....
	
	// different name, but exactly the same code!

UShort ccitt_updcrc_sega(UShort icrc, void* icp, unsigned short icnt)
{
#define M1 0xff
#define M2 0xff00
		
    register unsigned short crc = icrc;
    register unsigned char *cp = icp;
    register unsigned short cnt = icnt;
	register unsigned short *crctab;
	
	asm {
		lea		@crcTab, crctab
		moveq	#0,d1
		bra.s	@end

@nxt	move.w	crc,d0
		lsr.w	#8,d0
		move.b	(cp)+,d1
		eor.w	d1,d0
		add.w	d0,d0
		move.w	(crctab,d0),d0
		lsl.w	#8,crc
		eor.w	d0,crc

@end	dbra	cnt,@nxt
		}
		
#ifdef NEVER		
    while(cnt--) {
	crc=((crc<<8)&M2)^crctab[((crc>>8))^*cp++];
    }
#endif

    return(crc);

	asm {
@crcTab:	dc.w 0x0000
			dc.w 0x1021
			dc.w 0x2042
			dc.w 0x3063
			dc.w 0x4084
			dc.w 0x50a5
			dc.w 0x60c6
			dc.w 0x70e7
			dc.w 0x8108
			dc.w 0x9129
			dc.w 0xa14a
			dc.w 0xb16b
			dc.w 0xc18c
			dc.w 0xd1ad
			dc.w 0xe1ce
			dc.w 0xf1ef
			dc.w 0x1231
			dc.w 0x0210
			dc.w 0x3273
			dc.w 0x2252
			dc.w 0x52b5
			dc.w 0x4294
			dc.w 0x72f7
			dc.w 0x62d6
			dc.w 0x9339
			dc.w 0x8318
			dc.w 0xb37b
			dc.w 0xa35a
			dc.w 0xd3bd
			dc.w 0xc39c
			dc.w 0xf3ff
			dc.w 0xe3de
			dc.w 0x2462
			dc.w 0x3443
			dc.w 0x0420
			dc.w 0x1401
			dc.w 0x64e6
			dc.w 0x74c7
			dc.w 0x44a4
			dc.w 0x5485
			dc.w 0xa56a
			dc.w 0xb54b
			dc.w 0x8528
			dc.w 0x9509
			dc.w 0xe5ee
			dc.w 0xf5cf
			dc.w 0xc5ac
			dc.w 0xd58d
			dc.w 0x3653
			dc.w 0x2672
			dc.w 0x1611
			dc.w 0x0630
			dc.w 0x76d7
			dc.w 0x66f6
			dc.w 0x5695
			dc.w 0x46b4
			dc.w 0xb75b
			dc.w 0xa77a
			dc.w 0x9719
			dc.w 0x8738
			dc.w 0xf7df
			dc.w 0xe7fe
			dc.w 0xd79d
			dc.w 0xc7bc
			dc.w 0x48c4
			dc.w 0x58e5
			dc.w 0x6886
			dc.w 0x78a7
			dc.w 0x0840
			dc.w 0x1861
			dc.w 0x2802
			dc.w 0x3823
			dc.w 0xc9cc
			dc.w 0xd9ed
			dc.w 0xe98e
			dc.w 0xf9af
			dc.w 0x8948
			dc.w 0x9969
			dc.w 0xa90a
			dc.w 0xb92b
			dc.w 0x5af5
			dc.w 0x4ad4
			dc.w 0x7ab7
			dc.w 0x6a96
			dc.w 0x1a71
			dc.w 0x0a50
			dc.w 0x3a33
			dc.w 0x2a12
			dc.w 0xdbfd
			dc.w 0xcbdc
			dc.w 0xfbbf
			dc.w 0xeb9e
			dc.w 0x9b79
			dc.w 0x8b58
			dc.w 0xbb3b
			dc.w 0xab1a
			dc.w 0x6ca6
			dc.w 0x7c87
			dc.w 0x4ce4
			dc.w 0x5cc5
			dc.w 0x2c22
			dc.w 0x3c03
			dc.w 0x0c60
			dc.w 0x1c41
			dc.w 0xedae
			dc.w 0xfd8f
			dc.w 0xcdec
			dc.w 0xddcd
			dc.w 0xad2a
			dc.w 0xbd0b
			dc.w 0x8d68
			dc.w 0x9d49
			dc.w 0x7e97
			dc.w 0x6eb6
			dc.w 0x5ed5
			dc.w 0x4ef4
			dc.w 0x3e13
			dc.w 0x2e32
			dc.w 0x1e51
			dc.w 0x0e70
			dc.w 0xff9f
			dc.w 0xefbe
			dc.w 0xdfdd
			dc.w 0xcffc
			dc.w 0xbf1b
			dc.w 0xaf3a
			dc.w 0x9f59
			dc.w 0x8f78
			dc.w 0x9188
			dc.w 0x81a9
			dc.w 0xb1ca
			dc.w 0xa1eb
			dc.w 0xd10c
			dc.w 0xc12d
			dc.w 0xf14e
			dc.w 0xe16f
			dc.w 0x1080
			dc.w 0x00a1
			dc.w 0x30c2
			dc.w 0x20e3
			dc.w 0x5004
			dc.w 0x4025
			dc.w 0x7046
			dc.w 0x6067
			dc.w 0x83b9
			dc.w 0x9398
			dc.w 0xa3fb
			dc.w 0xb3da
			dc.w 0xc33d
			dc.w 0xd31c
			dc.w 0xe37f
			dc.w 0xf35e
			dc.w 0x02b1
			dc.w 0x1290
			dc.w 0x22f3
			dc.w 0x32d2
			dc.w 0x4235
			dc.w 0x5214
			dc.w 0x6277
			dc.w 0x7256
			dc.w 0xb5ea
			dc.w 0xa5cb
			dc.w 0x95a8
			dc.w 0x8589
			dc.w 0xf56e
			dc.w 0xe54f
			dc.w 0xd52c
			dc.w 0xc50d
			dc.w 0x34e2
			dc.w 0x24c3
			dc.w 0x14a0
			dc.w 0x0481
			dc.w 0x7466
			dc.w 0x6447
			dc.w 0x5424
			dc.w 0x4405
			dc.w 0xa7db
			dc.w 0xb7fa
			dc.w 0x8799
			dc.w 0x97b8
			dc.w 0xe75f
			dc.w 0xf77e
			dc.w 0xc71d
			dc.w 0xd73c
			dc.w 0x26d3
			dc.w 0x36f2
			dc.w 0x0691
			dc.w 0x16b0
			dc.w 0x6657
			dc.w 0x7676
			dc.w 0x4615
			dc.w 0x5634
			dc.w 0xd94c
			dc.w 0xc96d
			dc.w 0xf90e
			dc.w 0xe92f
			dc.w 0x99c8
			dc.w 0x89e9
			dc.w 0xb98a
			dc.w 0xa9ab
			dc.w 0x5844
			dc.w 0x4865
			dc.w 0x7806
			dc.w 0x6827
			dc.w 0x18c0
			dc.w 0x08e1
			dc.w 0x3882
			dc.w 0x28a3
			dc.w 0xcb7d
			dc.w 0xdb5c
			dc.w 0xeb3f
			dc.w 0xfb1e
			dc.w 0x8bf9
			dc.w 0x9bd8
			dc.w 0xabbb
			dc.w 0xbb9a
			dc.w 0x4a75
			dc.w 0x5a54
			dc.w 0x6a37
			dc.w 0x7a16
			dc.w 0x0af1
			dc.w 0x1ad0
			dc.w 0x2ab3
			dc.w 0x3a92
			dc.w 0xfd2e
			dc.w 0xed0f
			dc.w 0xdd6c
			dc.w 0xcd4d
			dc.w 0xbdaa
			dc.w 0xad8b
			dc.w 0x9de8
			dc.w 0x8dc9
			dc.w 0x7c26
			dc.w 0x6c07
			dc.w 0x5c64
			dc.w 0x4c45
			dc.w 0x3ca2
			dc.w 0x2c83
			dc.w 0x1ce0
			dc.w 0x0cc1
			dc.w 0xef1f
			dc.w 0xff3e
			dc.w 0xcf5d
			dc.w 0xdf7c
			dc.w 0xaf9b
			dc.w 0xbfba
			dc.w 0x8fd9
			dc.w 0x9ff8
			dc.w 0x6e17
			dc.w 0x7e36
			dc.w 0x4e55
			dc.w 0x5e74
			dc.w 0x2e93
			dc.w 0x3eb2
			dc.w 0x0ed1
			dc.w 0x1ef0
			};
}
